/* gensagdata.c --- Generate C code with Sagittal data from
   `CharMap2.txt' and `NumDen2.txt' */

/* This program is part of microabc (C) 2006 Hudson Lacerda */

/* Copyright  (C)  2006  Hudson F. M. Lacerda <hfmlacerda @ yahoo.com.br> */

/* Author: Hudson F. M. Lacerda <hfmlacerda @ yahoo.com.br> */
/* Maintainer: Hudson F. M. Lacerda <hfmlacerda @ yahoo.com.br> */
/* URL: http://br.geocities.com/hfmlacerda/ */

/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 2 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software */
/* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA */


#include <stdio.h>
#include <stdlib.h>
#include <math.h> /* compile/link with gcc -lm */

/* file names */
const char cm_name[] = "CharMap2.txt";
const char nd_name[] = "NumDen2.txt";
const char thisprogram[] = "gensagdata.c";
const char outfile[] = "sagdata.c";

static char *fgetline(char s[],int lim,FILE *ap_arquivo);
static int iseol(char c);
static double interv2cents(double interv);
static double cents2interv(double cents);
static int filtstr(char *s, int lim);
int read_charmap(FILE *fp);
int read_numden(FILE *fp);
int output_struct(FILE *fp);
int initialize_sagittal();


/* WARNING: This MUST be the SAME struct printed by output_struct */
/* You need to create sagdata.h with the data below */

/* begin sagdata.h */

#define NAMESIZE (16) /* max size of an ASCII representation of a
			 Sagittal accidental */
#define NSAGCHARS (256) /* number of characters (accidentals) data to
			   store (size of sagittal font) */
typedef struct {
    int code;
    double num;
    double den;
    double cents;
    char shortpure[NAMESIZE];
    char longpure[NAMESIZE];
    char shortmixed[NAMESIZE];
    char longmixed[NAMESIZE];
    char alternative[NAMESIZE];
} sagstruct;

/* end sagdata.h */

static int sagcpycm( sagstruct *dest, sagstruct *orig);
static int sagcpynd( sagstruct *dest, sagstruct *orig);

static sagstruct sagittal[NSAGCHARS];

int main()
{

    FILE *cm, *nd;

    /* initialize structs vector */
    initialize_sagittal();

    /* read char map */
    if((cm=fopen(cm_name,"rt"))==NULL) {
	fprintf(stderr,"ERROR: Can't open file \"%s\"\n",cm_name);
       exit(-1);
    }
    read_charmap(cm);
    fclose(cm);

    /* read num den */
    if((nd=fopen(nd_name,"rt"))==NULL) {
       fprintf(stderr,"ERROR: Can't open file \"%s\"\n",nd_name);
       exit(-1);
    }
    read_numden(nd);
    fclose(nd);

    /* output struct */
    output_struct(stdout);

}

char *fgetline(char s[],int lim,FILE *ap_arquivo)
{
    int c,i;

    s[0]='\0';    
    for(i=0; i<lim-1
	    && (c=fgetc(ap_arquivo))!=EOF
	    && (!iseol(c));i++)
	s[i]=c;
    if(iseol(c)) {
	s[i]=c;
	i++;
    }
    s[i]='\0';

    if(c==EOF && strlen(s)==0)
	return NULL;

    return s;
}

int iseol(char c)
{
    switch(c) {
    case 10:
    case 13:
	return 1;
    default:
	return 0;
    }
}

static double interv2cents(double interv)
{
    return 1200.0 * (log(interv)/log(2.0));
}

static double cents2interv(double cents)
{
    return pow(2.0,cents/1200.0);
}

int read_charmap(FILE *fp)
{
    char line[BUFSIZ];
    sagstruct sag;
    while (fgetline(line,BUFSIZ,fp)!=NULL) {
	if(sscanf(line,"%d %s %s %s %s",
		  &sag.code,
		  &sag.shortpure,
		  &sag.longpure,
		  &sag.shortmixed,
		  &sag.longmixed)
	   == 5 ) {
	    /* treat backslashes and double quotes */
	    filtstr(sag.shortpure,NAMESIZE);
	    filtstr(sag.longpure,NAMESIZE);
	    filtstr(sag.shortmixed,NAMESIZE);
	    filtstr(sag.longmixed,NAMESIZE);

	    //
	    /* TODO: generate alternative*/
	    //
	    //filtstr(sag.alternative,NAMESIZE);

	    /* copy data to sagittal struct */
	    sagcpycm(&sagittal[sag.code],&sag);
	}
    }
}

int read_numden(FILE *fp)
{
    char line[BUFSIZ];
    sagstruct sag;
    while (fgetline(line,BUFSIZ,fp)!=NULL) {
	if(sscanf(line,"%d %lf %lf",
		  &sag.code,
		  &sag.num,
		  &sag.den)
	   == 3 ) {
	    /* compute cents */
	    sag.cents = interv2cents(sag.num/sag.den);
	    /* copy data to sagittal struct */
	    sagcpynd(&sagittal[sag.code],&sag);
	}
    }
}


static int sagcpycm(sagstruct *dest,sagstruct *orig)
{
    strcpy(dest->shortpure, orig->shortpure);
    strcpy(dest->longpure, orig->longpure);
    strcpy(dest->shortmixed, orig->shortmixed);
    strcpy(dest->longmixed, orig->longmixed);
    strcpy(dest->alternative, orig->alternative);
}

static int sagcpynd(sagstruct *dest,sagstruct *orig)
{
    /* dest->code = orig->code ; */
    dest->num = orig->num ;
    dest->den = orig->den ;
    dest->cents = orig->cents ;
}

int initialize_sagittal()
{
    int i;
    for(i=0;i<256;i++) {

	sagittal[i].code = i ;

	sagittal[i].num=1.0;
	sagittal[i].den=1.0;
	sagittal[i].cents=0.0;
	sagittal[i].shortpure[0]='\0';
	sagittal[i].longpure[0]='\0';
	sagittal[i].shortmixed[0]='\0';
	sagittal[i].longmixed[0]='\0';
	sagittal[i].alternative[0]='\0';
    }
}

int output_struct(FILE *fp)
{
    int i, b, e;
    b=0; /*begin*/
    e=255; /*end*/
    fprintf (fp, "/* %s - Automatically generated by %s (Don't edit) */\n",
	     outfile,thisprogram);

    fprintf(fp,"\n#include \"sagdata.h\"\n");

    fprintf(fp, "\nsagstruct sagittal[%d] ={\n",NSAGCHARS);

    for(i=b;i<=e;i++) {
	fprintf(fp, "\t{ %d,", sagittal[i].code );
	fprintf(fp, "%.2lf,", sagittal[i].num );
	fprintf(fp, "%.2lf,", sagittal[i].den );
	fprintf(fp, "%.20lf,", sagittal[i].cents );
	fprintf(fp, "\"%s\",", sagittal[i].shortpure );
	fprintf(fp, "\"%s\",", sagittal[i].longpure );
	fprintf(fp, "\"%s\",", sagittal[i].shortmixed );
	fprintf(fp, "\"%s\",", sagittal[i].longmixed );
	fprintf(fp, "\"%s\"", sagittal[i].alternative );
	if(i<e)
	    fprintf(fp," },\n");
    }

    fprintf(fp, " }\n};\n\n/* end of file %s */\n",outfile);

}

static int filtstr(char *s, int lim) /* escape \ and " in s[] */
{
    int n;
    char t[BUFSIZ];
    char c;
    char *p;
    p=s;
    n=0;
    while((c=(*p))!='\0' && n<lim) {
	if(c=='\\')
	    t[n++]=c;
	else if(c=='"')
	    t[n++]='\\';
	t[n++]=c;
	   p++;
    }
    t[n]='\0';
    strncpy(s,t,lim);
}

